/*------------------------------------------------------------------------------
	Newton interface to the native image decoding functions.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "NewtonScript.h"
#include "Newton/ObjectSystem.h"
#include "Newton/NewtonQD.h"

#define	kGestalt_SystemInfo	0x01000003

/*------------------------------------------------------------------------------
	Return an options frame sutable for passing to MakeBitmap.
	Check the actual screen depth.

	Args:		inStore
					the store on which to create the bitmap object

	Return:	the options frame ref
------------------------------------------------------------------------------*/

Ref
MakeOptionsFrame(RefArg inStore)
{
	RefVar options(AllocateFrame());

#if defined(__NCT__)
	RefVar gestalt(NSCallGlobalFn(SYM(Gestalt), MAKEINT(kGestalt_SystemInfo)));
	RefVar screenDepth(GetFrameSlot(gestalt, SYM(screenDepth)));
	if (ISNIL(screenDepth))
		screenDepth = MAKEINT(4);
#else
	RefVar screenDepth(MAKEINT(4));  // Newton.framework doesnt do Gestalt yet
#endif
	SetFrameSlot(options, SYM(depth), screenDepth);
	if (NOTNIL(inStore))
		SetFrameSlot(options, SYM(store), inStore);

	return options;
}


#if !defined(__NCT__)
/*------------------------------------------------------------------------------
	Stubbed functions for the Project Builder emulation.

	None needed any more - its all handled by the Newton.framework!
------------------------------------------------------------------------------*/

#endif

#pragma mark -

/*------------------------------------------------------------------------------
	Newton interface to the native GIF decoding function.
------------------------------------------------------------------------------*/

#include "GIF.h"

/*------------------------------------------------------------------------------
	Decode GIF image.

	Args:		inData
					the GIF data
				inStore
					store on which to create VBO

	Return:	a bitmap binary object
------------------------------------------------------------------------------*/

extern "C"
Ref DecodeGIF(RefArg rcvr, RefArg inData, RefArg inStore)
{
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);

	RefVar bitmap;
	RefVar options(MakeOptionsFrame(inStore));
	CGIFReader * reader = NULL;

	WITH_LOCKED_BINARY(inData, dataPtr)
		UShort width, height;
		reader = new CGIFReader((UChar *) dataPtr, Length(inData));
		if (reader->StartGIF(&width, &height))
		{
			bitmap = NSCallGlobalFn(SYM(MakeBitmap), MAKEINT(width), MAKEINT(height), options);
			if (NOTNIL(bitmap))
			{
				RefVar pix(GetFrameSlot(bitmap, SYM(data)));
				WITH_LOCKED_BINARY(pix, pixPtr)
					reader->ReadGIF((PixelMap *) pixPtr);
				END_WITH_LOCKED_BINARY(pix)
			}
		}
/*	END_WITH_LOCKED_BINARY(inData)
	 but we also ALWAYS want to delete the reader so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
	}
	on_unwind {
		UnlockRefArg(inData);
		if (reader)
			delete reader;
	}
	end_unwind;

	return bitmap;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Newton interface to the native JPEG decoding function.
------------------------------------------------------------------------------*/

#include "jinclude.h"
#include "jpeglib.h"

EXTERN(void) jpeg_newton_src JPP((j_decompress_ptr cinfo, JOCTET * dataPtr, size_t length));

/*------------------------------------------------------------------------------
	Decode JPEG image.

	Args:		inData
					the JPEG data
				inStore
					store on which to create VBO

	Return:	a bitmap binary object
------------------------------------------------------------------------------*/

extern "C"
Ref DecodeJPEG(RefArg rcvr, RefArg inData, RefArg inStore)
{
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);

	RefVar bitmap;
	RefVar options(MakeOptionsFrame(inStore));

	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;

	/* Initialize the JPEG decompression object with Newton (NULL) error handling. */
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	/* Insert custom marker processor for COM and APP12.
	* APP12 is used by some digital camera makers for textual info,
	* so we provide the ability to display it as text.
	* If you like, additional APPn marker types can be selected for display,
	* but don't try to override APP0 or APP14 this way (see libjpeg.doc).
	jpeg_set_marker_processor(&cinfo, JPEG_COM, print_text_marker);
	jpeg_set_marker_processor(&cinfo, JPEG_APP0+12, print_text_marker);
	*/

	WITH_LOCKED_BINARY(inData, dataPtr)

		/* Specify data source for decompression */
		jpeg_newton_src(&cinfo, (JOCTET *) dataPtr, (size_t) Length(inData));

		/* Read file header, set default decompression parameters */
		jpeg_read_header(&cinfo, TRUE);

		/* Adjust default decompression parameters */
		/* Set up to generate greyscale image */
		cinfo.out_color_space = JCS_GRAYSCALE;

		/* Start decompressor */
		jpeg_start_decompress(&cinfo);

		bitmap = NSCallGlobalFn(SYM(MakeBitmap), MAKEINT(cinfo.output_width),
															  MAKEINT(cinfo.output_height), options);
		if (NOTNIL(bitmap))
		{
			RefVar pix(GetFrameSlot(bitmap, SYM(data)));
			WITH_LOCKED_BINARY(pix, pixPtr)
				PixelMap * pixmap = (PixelMap *) pixPtr;
				Ptr pixBits = GetPixelMapBits(pixmap);
				int pixRowBytes = pixmap->rowBytes & 0x3FFF;

				// Allocate space for the decode buffer
				JSAMPARRAY image_buffer = (*cinfo.mem->alloc_sarray)
					((j_common_ptr) &cinfo, JPOOL_IMAGE, cinfo.output_width, (JDIMENSION) cinfo.rec_outbuf_height);

				/* Process data */
				while (cinfo.output_scanline < cinfo.output_height)
				{
					int rowNumber = cinfo.output_scanline;
					JDIMENSION num_scanlines = jpeg_read_scanlines(&cinfo, image_buffer, cinfo.rec_outbuf_height);
					while (num_scanlines-- > 0)
					{
						UChar * srcPtr = image_buffer[0];
						UChar * dstPtr = (UChar *) pixBits + pixRowBytes * rowNumber++;
						UChar pixHi, pixLo;
						int i;
						for (i = cinfo.output_width/2; i > 0; i--)
						{
							pixHi = *srcPtr++ & 0xF0;
							pixLo = *srcPtr++ >> 4;
							*dstPtr++ = ~(pixHi | pixLo);
						}
						if (cinfo.output_width & 1)	// if width is odd, write out odd pixel
							*dstPtr = ~(*srcPtr++ & 0xF0);
					}
				}
			END_WITH_LOCKED_BINARY(pix)
		}

/*	END_WITH_LOCKED_BINARY(inData)
	 but we also ALWAYS want to delete the decompressor so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
	}
	on_unwind {
		UnlockRefArg(inData);
		jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
	}
	end_unwind;

	return bitmap;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Newton interface to the native TIFF decoding function.
------------------------------------------------------------------------------*/

#include "tiffiop.h"

extern "C" int TIFFPixImageGet(TIFFRGBAImage * img, Ptr pixBits, uint32 rowBytes, uint32 wd, uint32 ht);

/*------------------------------------------------------------------------------
	File-oriented interface to libtiff.
	Wrapped in a class whose pointer can be passed around instead of
	a file descriptor.
	Ugly.
------------------------------------------------------------------------------*/

class XFile
{
public:
						XFile(void * data, size_t size);
	static tsize_t	read(thandle_t, tdata_t buf, tsize_t size);
	static tsize_t	write(thandle_t, tdata_t buf, tsize_t size);
	static toff_t	size(thandle_t);
	static toff_t	seek(thandle_t, toff_t off, int whence);
	static int		map(thandle_t, tdata_t * pbase, toff_t * psize);
	static void		unmap(thandle_t, tdata_t base, toff_t size);
	static int		close(thandle_t);

private:
// TIFF input stream
	UChar *		dataPtr;
	UChar *		dataStart;
	size_t		dataSize;

// bitmap output stream
	PixelMap *	pixmap;
	Ptr			pixBits;
	size_t		pixRowBytes;
};


/*------------------------------------------------------------------------------
	File-oriented interface to tifflib.
------------------------------------------------------------------------------*/

XFile::XFile(void * data, size_t size)
{
	dataPtr = dataStart = (UChar *) data;
	dataSize = size;
}

toff_t
XFile::size(thandle_t context)
{
	return ((XFile *) context)->dataSize;
}

toff_t
XFile::seek(thandle_t context, toff_t off, int whence)
{
	XFile * self = (XFile *) context;
	long fpos = self->dataPtr - self->dataStart;

	switch (whence)
	{
	case SEEK_CUR:
		if (off + fpos > self->dataSize)
			return EOF;		// can't extend
		self->dataPtr += off;
		break;
	case SEEK_END:
		if (off > 0)
			return EOF;		// can't extend
		self->dataPtr = self->dataStart + self->dataSize + off;
		break;
	case SEEK_SET:
		if (off > self->dataSize)
			return EOF;		// can't extend
		self->dataPtr = self->dataStart + off;
		break;
	}

	return (toff_t) (self->dataPtr - self->dataStart);
}

tsize_t
XFile::read(thandle_t context, tdata_t buf, tsize_t size)
{
	XFile * self = (XFile *) context;
	tsize_t sizeRemaining = self->dataSize - (self->dataPtr - self->dataStart);
	if (size > sizeRemaining)
		size = sizeRemaining;
	if (size > 0)
	{
		memcpy(buf, self->dataPtr, size);
		self->dataPtr += size;
	}
	return size;
}

tsize_t
XFile::write(thandle_t context, tdata_t buf, tsize_t size)
{
	return 0;
}

int
XFile::map(thandle_t context, tdata_t * pbase, toff_t * psize)
{
	return 0;
}

void
XFile::unmap(thandle_t context, tdata_t base, toff_t size)
{
}

int
XFile::close(thandle_t context)
{
	return 0;
}


/*------------------------------------------------------------------------------
	Memory interface to libtiff.
------------------------------------------------------------------------------*/

void
_TIFFmemset(tdata_t p, int v, tsize_t c)
{
	memset(p, v, (size_t) c);
}

void
_TIFFmemcpy(tdata_t d, const tdata_t s, tsize_t c)
{
	memcpy(d, s, (size_t) c);
}

int
_TIFFmemcmp(const tdata_t p1, const tdata_t p2, tsize_t c)
{
	return memcmp(p1, p2, (size_t) c);
}

tdata_t
_TIFFmalloc(tsize_t s)
{
	return NewPtr((size_t) s);
}

void
_TIFFfree(tdata_t p)
{
	DisposePtr((Ptr) p);
}

tdata_t
_TIFFrealloc(tdata_t p, tsize_t s)
{
	return ReallocPtr((Ptr) p, (size_t) s);
}


/*------------------------------------------------------------------------------
	Decode TIFF image.

	Args:		inData
					the TIFF data
				inStore
					store on which to create VBO

	Return:	a bitmap binary object
------------------------------------------------------------------------------*/

extern "C"
Ref DecodeTIFF(RefArg rcvr, RefArg inData, RefArg inStore)
{
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);

	RefVar bitmap;
	RefVar options(MakeOptionsFrame(inStore));

	XFile * x = NULL;
	TIFF * tif = NULL;
	char * emsg = NULL;
	WITH_LOCKED_BINARY(inData, dataPtr)
		// instantiate TIFF decoder
		x = new XFile(dataPtr, Length(inData));
		tif = TIFFClientOpen("Newton", "r", (thandle_t) x,
									(TIFFReadWriteProc) x->read,
									(TIFFReadWriteProc) x->write,
									(TIFFSeekProc) x->seek,
									(TIFFCloseProc) x->close,
									(TIFFSizeProc) x->size,
									(TIFFMapFileProc) x->map,
									(TIFFUnmapFileProc) x->unmap);
		if (tif)
		{
			TIFFRGBAImage img;
			emsg = new char[256];
	
			if (TIFFRGBAImageBegin(&img, tif, NO, emsg))
			{
				SetFrameSlot(options, SYM(depth), MAKEINT(img.bitspersample > 1 ? 4 : 1));
				bitmap = NSCallGlobalFn(SYM(MakeBitmap), MAKEINT(img.width), MAKEINT(img.height), options);
				if (NOTNIL(bitmap))
				{
					RefVar pix(GetFrameSlot(bitmap, SYM(data)));
					WITH_LOCKED_BINARY(pix, pixPtr)
						PixelMap * pixmap = (PixelMap *) pixPtr;
						Ptr pixBits = GetPixelMapBits(pixmap);
						int pixRowBytes = pixmap->rowBytes & 0x3FFF;

					// decode TIFF
						TIFFPixImageGet(&img, pixBits, pixRowBytes, pixmap->bounds.right, pixmap->bounds.bottom);

					// delete TIFF decoder
						TIFFRGBAImageEnd(&img);
					END_WITH_LOCKED_BINARY(pix)
				}
			}
			else
				TIFFError("Decode", emsg);

			TIFFClose(tif), tif = NULL;
			delete[] emsg, emsg = NULL;
		}
		delete x, x = NULL;

/*	END_WITH_LOCKED_BINARY(inData)
	 but we also ALWAYS want to delete the decompressor so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
	}
	on_unwind {
		UnlockRefArg(inData);
		if (tif)
			_TIFFfree(tif);
		if (emsg)
			delete[] emsg;
		if (x)
			delete x;
	}
	end_unwind;

	return bitmap;
}

#pragma mark -

/*------------------------------------------------------------------------------
	Newton interface to the native PNG decoding function.
------------------------------------------------------------------------------*/

#include "png.h"

/*------------------------------------------------------------------------------
	Provide an error exit.
------------------------------------------------------------------------------*/

extern "C"
void newt_error(png_structp png_ptr, png_const_charp msg)
{
#if !defined(__NCT__)
	printf(msg);
#endif
	ThrowMsg((char *) msg);
}

extern "C"
void newt_warning(png_structp png_ptr, png_const_charp msg)
{
#if !defined(__NCT__)
	printf(msg);
#endif
}


/*------------------------------------------------------------------------------
	I/O functions.
------------------------------------------------------------------------------*/

typedef struct
{
	char *	dataStart;
	char *	dataStop;
	char *	dataPtr;
} SReadContext;

static void
newt_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	SReadContext * context = (SReadContext *) png_get_io_ptr(png_ptr);
	png_size_t lengthAvailable = context->dataStop - context->dataPtr;

	if (length > lengthAvailable)
		length = lengthAvailable;
	if (length)
	{
		memcpy(data, context->dataPtr, length);
		context->dataPtr += length;
	}
}

#if 0
static void
newt_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	SReadContext * context = (SReadContext *) png_get_io_ptr(png_ptr);
	png_size_t lengthAvailable = context->dataStop - context->dataPtr;

	if (length > lengthAvailable)
		length = lengthAvailable;
	if (length)
	{
		memcpy(context->dataPtr, data, length);
		context->dataPtr += length;
	}
}

static void
newt_flush(png_structp png_ptr)
{
//	No need to flush.
}
#endif


/*------------------------------------------------------------------------------
	Decode PNG image.

	Args:		inData
					the PNG data
				inStore
					store on which to create VBO

	Return:	a bitmap binary object
------------------------------------------------------------------------------*/

extern "C"
Ref DecodePNG(RefArg rcvr, RefArg inData, RefArg inStore)
{
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);

	RefVar bitmap;
	RefVar options(MakeOptionsFrame(inStore));

	png_structp		png_ptr;
	png_infop		info_ptr;
	png_uint_32		imageWidth, imageHeight, row;
	int				bit_depth, color_type, interlace_type;
	int				pass, numOfPasses;
	SReadContext	newt_io;

   /* Create and initialize the png_struct with the desired error handler
    * functions.  We also supply the compiler header file version, so that
	 * we know if the application was compiled with a compatible version of
	 * the library.
    */
   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
												NULL, newt_error, newt_warning);
   if (png_ptr == NULL)
      return NILREF;

   /* Allocate/initialize the memory for image information. */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
      png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
      return NILREF;
   }

	WITH_LOCKED_BINARY(inData, dataPtr)

		newt_io.dataStart = (char *) dataPtr;
		newt_io.dataStop = (char *) dataPtr + Length(inData);
		newt_io.dataPtr = (char *) dataPtr;
		png_set_read_fn(png_ptr, (void *) &newt_io, newt_read_data);

		/* The call to png_read_info() gives us all of the information from the
		* PNG file before the first IDAT (image data chunk).
		*/
		png_read_info(png_ptr, info_ptr);

		png_get_IHDR(png_ptr, info_ptr,
						&imageWidth, &imageHeight, &bit_depth,
						&color_type, &interlace_type, int_p_NULL, int_p_NULL);

		/* Set up the data transformations you want.  Note that these are all
		* optional.  Only call them if you want/need them.  Many of the
		* transformations only work on specific types of images, and many
		* are mutually exclusive.
		*/

		/* Tell libpng to strip 16 bit/color files down to 8 bits/color. */
		png_set_strip_16(png_ptr);

		/* Strip any alpha channel. */
		png_set_strip_alpha(png_ptr);

		/* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
		* byte into separate bytes (useful for paletted and grayscale images).
		*/
		png_set_packing(png_ptr);

		/* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
		*/
		if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
			png_set_gray_1_2_4_to_8(png_ptr);

		/* Transform RGB to greyscale.
		*/
		png_set_rgb_to_gray(png_ptr, 1, -1, -1);

		/* Turn on interlace handling.  REQUIRED */
		numOfPasses = png_set_interlace_handling(png_ptr);

		/* Optional call to gamma correct and add the background to the palette
		* and update info structure.  REQUIRED if you are expecting libpng to
		* update the palette for you (ie you selected such a transform above).
		*/
		png_read_update_info(png_ptr, info_ptr);

		/* Create a PixelMap large enough to receive the image */
		bitmap = NSCallGlobalFn(SYM(MakeBitmap), MAKEINT(imageWidth),
															  MAKEINT(imageHeight), options);
		if (NOTNIL(bitmap))
		{
			png_bytep rowPtr = NULL;
			RefVar pix(GetFrameSlot(bitmap, SYM(data)));
			WITH_LOCKED_BINARY(pix, pixPtr)
				PixelMap * pixmap = (PixelMap *) pixPtr;
				Ptr pixBits = GetPixelMapBits(pixmap);
				int pixRowBytes = pixmap->rowBytes & 0x3FFF;

				// Allocate space for the decode buffer
				rowPtr = (png_bytep) png_malloc(png_ptr, png_get_rowbytes(png_ptr, info_ptr));
		
				/* Read the image one scanline at a time - deal with interlacing */
				for (pass = 0; pass < numOfPasses; pass++)
				{
					for (row = 0; row < imageHeight; row++)
					{
						png_read_row(png_ptr, rowPtr, png_bytep_NULL);
		
						UChar * srcPtr = rowPtr;
						UChar * dstPtr = (UChar *) pixBits + pixRowBytes * row;
						UChar pixHi, pixLo;
						int i;
						for (i = imageWidth/2; i > 0; i--)
						{
							pixHi = *srcPtr++ & 0xF0;
							pixLo = *srcPtr++ >> 4;
							*dstPtr++ = ~(pixHi | pixLo);
						}
						if (imageWidth & 1)	// if width is odd, write out odd pixel
							*dstPtr = ~(*srcPtr++ & 0xF0);
					}
				}
/*			END_WITH_LOCKED_BINARY(pix)
			 but we also ALWAYS want to free the row bufer so we expand the macro and include the delete */
#if 0
			{	// balance bracket for source code parser
#endif
			}
			on_unwind {
				UnlockRefArg(pix);
				if (rowPtr)
					png_free(png_ptr, rowPtr);
				png_read_end(png_ptr, info_ptr);
			}
			end_unwind;
		}

/*	END_WITH_LOCKED_BINARY(inData)
	 but we also ALWAYS want to delete the decompressor so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
	}
	on_unwind {
		UnlockRefArg(inData);
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
	}
	end_unwind;

	return bitmap;
}

#pragma mark -

#if defined(__NCT__)
/*------------------------------------------------------------------------------
	Stubs to keep the linker happy.
	The TIFF library (unlike the pleasant PNG library) does not allow
	FILE and jmp_buf system stuff to be ignored.
------------------------------------------------------------------------------*/

extern "C" int _sprintf(char * s, const char * fmt, ...)
{	return 0; }

extern "C" int fprintf(FILE * f, const char * fmt, ...)
{	return 0; }

extern "C" void __assertprint(const char * file, int line, const char * expr)
{}

/* setjmp/longjmp is used by tif_jpeg.c to catch libjpeg errors.
 * This should be sorted out if we really care about error handling.
*/
#include <setjmp.h>
extern "C" int setjmp(jmp_buf env)
{	return 0; }

extern "C" void longjmp(jmp_buf env, int val)
{}
#endif
