
#if !defined(__GIF_H)
#define __GIF_H 1

#include <stdlib.h>

#if !defined(__NEWTONQD_H)
#include "Newton/NewtonQD.h"
#endif

#if defined(__APPLE_CC__)
#define ThrowBadTypeWithFrameData(error, value)
#else
#define kRefTagBits		2
#define	MAKEINT(i)		(((long) (i)) << kRefTagBits)
#endif


/*----------------------------------------------------------------------
	Maximum number of LZW codes
----------------------------------------------------------------------*/

#define GIFBITS		  12

#define MAX_CODES		(1 << GIFBITS)
#define HASH_SIZE		5003		// for 80% code hash table occupancy
#define HASH_BITS			4


/*----------------------------------------------------------------------
	GIF byte codes.
----------------------------------------------------------------------*/

#define kExtensionIntro			0x21
#define kBlockTerminator		0x00
#define kTrailer					0x3B


/*----------------------------------------------------------------------
	Standard Red/Green/Blue colour structure.
----------------------------------------------------------------------*/

typedef struct
{
	UByte	r, g, b, u;
} GIFRGB, * GIFRGBPtr, ** GIFRGBHandle;

	
/*----------------------------------------------------------------------
	Error Red/Green/Blue colour structure.
----------------------------------------------------------------------*/

typedef struct
{
	SShort	r, g, b, u;
} ERRGB, * ERRGBPtr, ** ERRGBHandle;


/*----------------------------------------------------------------------
	Logical Screen Descriptor.

	The Logical Screen Descriptor contains the parameters necessary to
	define the area of the display device within which the images will be rendered.
	The coordinates in this block are given with respect to the top-left corner
	of the virtual screen; they do not necessarily refer to absolute coordinates
	on the display device. This implies that they could refer to window coordinates in a
	window-based environment or printer coordinates when a printer is used. 
----------------------------------------------------------------------*/

typedef struct
{
	UShort		width;					// original fields from Data Stream
	UShort		height;
	UByte			flags;
	UByte			backgroundColor;
	UByte			aspectRatio;

	BOOL			hasGlobalColorTable;	// unpacked flags
	UByte			colorResolution;
	BOOL			isSorted;
	UShort		globalColorTableSize;

	GIFRGBPtr	globalColorTable;		// useful pointer, not part of LSD
} GIFLogicalScreenDescriptor, * GIFLogicalScreenDescriptorPtr,
									  ** GIFLogicalScreenDescriptorHandle;


/*----------------------------------------------------------------------
	Image descriptor.

	Each image in the Data Stream is composed of an Image Descriptor,
	an optional Local Color Table, and the image data. Each image must fit within
	the boundaries of the Logical Screen, as defined in the Logical Screen Descriptor.

	The Image Descriptor contains the parameters necessary to process a table based image.
	The coordinates given in this block refer to coordinates within the Logical Screen,
	and are given in pixels. This block is a Graphic-Rendering Block, optionally preceded
	by one or more Control blocks such as the Graphic Control Extension, and may be
	optionally followed by a Local Color Table; the Image Descriptor is always followed
	by the image data. 
----------------------------------------------------------------------*/

#define kImageLabel	0x2C

typedef struct
{
	UShort		left, top, width, height;	// original fields from Data Stream
	UByte			flags;

	BOOL			hasLocalColorTable;		// unpacked flags
	BOOL			isInterlaced;
	BOOL			isSorted;
	UShort		localColorTableSize;

	GIFRGBPtr	localColorTable;		//	useful pointers
	UByte *		data;
} GIFImageDescriptor, * GIFImageDescriptorPtr,
							** GIFImageDescriptorHandle;


/*----------------------------------------------------------------------
	Graphic Control Extension.

	The Graphic Control Extension contains parameters used when processing a
	graphic rendering block. The scope of this extension is the first
	graphic rendering block to follow.
----------------------------------------------------------------------*/

#define kGraphicControlLabel	0xF9
#define kGraphicControlBlockSize 4

typedef struct
{
	UByte			flags;					// original fields from Data Stream
	UShort		delayTime;
	UByte			transparentColor;

	BOOL			hasTransparency;		// unpacked flags
} GIFGraphicControlExtension, * GIFGraphicControlExtensionPtr,
									  ** GIFGraphicControlExtensionHandle;


/*----------------------------------------------------------------------
	Comment Extension.

	The Comment Extension contains textual information which is not part of the actual
	graphics in the GIF Data Stream. It is suitable for including comments about the
	graphics, credits, descriptions or any other type of non-control and non-graphic data.
	The Comment Extension may be ignored by the decoder, or it may be saved for later
	processing; under no circumstances should a Comment Extension disrupt or interfere
	with the processing of the Data Stream. 
----------------------------------------------------------------------*/

#define kCommentLabel 0xFE

typedef struct
{
	UByte *		data;
} GIFCommentExtension;


/*----------------------------------------------------------------------
	Plain Text Extension.

	The Plain Text Extension contains textual data and the parameters necessary to
	render that data as a graphic, in a simple form. The textual data will be encoded
	with the 7-bit printable ASCII characters. Text data are rendered using a grid of
	character cells defined by the parameters in the block fields. Each character is
	rendered in an individual cell. The textual data in this block is to be rendered
	as mono-spaced characters, one character per cell, with a best fitting font and size.
----------------------------------------------------------------------*/

#define kPlainTextLabel   0x01
#define kPlainTextBlockSize 12

typedef struct
{
	UShort		gridLeft, gridTop, gridWidth, gridHeight;
	UByte			cellWidth, cellHeight;
	UByte			foregroundColor, backgroundColor;
	UByte *		data;
} GIFPlainTextExtension;


/*----------------------------------------------------------------------
	Application Extension.

	The Application Extension contains application-specific information.
----------------------------------------------------------------------*/

#define kApplicationLabel   0xFF
#define kApplicationBlockSize 11

typedef struct
{
	UByte			id[8];
	UByte			auth[3];
	UByte *		data;
} GIFApplicationExtension;


/*----------------------------------------------------------------------
	C G I F R e a d e r
  --------------------------------------------------------------------
	GIF reader object.
----------------------------------------------------------------------*/

class CGIFReader
{
public:
				CGIFReader(UChar * data, size_t length);
				~CGIFReader();

	BOOL		StartGIF(UShort * width, UShort * height);
	void		ReadGIF(PixelMap * pixels);

private:
// GIF block data
	BOOL		ReadHeader();
	void		ReadLogicalScreenDescriptor();
	void		ReadImageDescriptor();
	BOOL		ReadGraphicControlExtension();
	void		ReadColorTable(UShort count, GIFRGBPtr ct);    
	void		TrashDataSubBlocks(void);
	void		TrashCommentExtension(void);
	void		TrashApplicationExtension(void);
	void		TrashPlainTextExtension(void);

	GIFLogicalScreenDescriptor	screen;
	GIFImageDescriptor			image;
	GIFGraphicControlExtension	gext;
	GIFRGBPtr	currentColorTable;
	BOOL		isInterlaced;
	size_t	bitsPerPixel;
	int		lineCount;
	int		pass;

// GIF input stream
	UChar		GetByte(void);
	UShort	GetWord(void);
	void		Advance(int count);
	UChar		GetBlockByte(void);
	ULong		GetCode(void);

	UChar *	dataPtr;
	UChar *	dataStart;
	UChar *	dataStop;
	size_t	availableBytes;			/* # bytes left in block */
	size_t	availableBits;				/* # bits left in current byte */
	UChar		currentByte;				/* current byte */

// LZW decoder
	void		InitDecoder(size_t size);
	void		ResetDecoder(void);
	BOOL		Decode(UShort linewidth);

	size_t	initialCodeSize;
	size_t	currentCodeSize;
	ULong		clearCode;
	ULong		endCode;
	ULong		newCodes;
	ULong		topSlot;
	ULong		slot;
	UChar		stack[MAX_CODES];		/* Stack for storing pixels */
	UChar		suffix[MAX_CODES];	/* Suffix table */
	UShort	prefix[MAX_CODES];	/* Prefix linked list */ 

// bitmap output stream
	void		OutputLine(UByte * line, size_t length);
	static void		Render1bit(UByte * line, size_t length, GIFRGBPtr ct, UByte * pixPtr);
	static void		Render4bit(UByte * line, size_t length, GIFRGBPtr ct, UByte * pixPtr);

	void			(*render)(UByte * line, size_t length, GIFRGBPtr ct, UByte * pixPtr);
	PixelMap *	pixmap;
	Ptr			pixBits;
	size_t		pixRowBytes;
};
	

/*----------------------------------------------------------------------
	C G I F W r i t e r
  --------------------------------------------------------------------
	GIF writer object.
----------------------------------------------------------------------*/

class CGIFWriter
{
public:
				CGIFWriter();
				~CGIFWriter();

	size_t	WriteGIF(const PixelMap * pixels, UChar * data, size_t length);

private:
// bitmap input stream
	int		GetNextPixel(void);

	const PixelMap *	pixmap;
	Ptr		pixBits;
	size_t	bitsPerPixel;
	size_t	resolution;
	size_t	pixRowBytes;
	int		pixWidth;
	int		pixHeight;
	Point		pixLoc;
	int		pass;
	BOOL		isInterlaced;
//	UChar		backgroundColor;
//	UChar		transparentColor;

// LZW encoder
	void		InitEncoder(size_t size);
	void		ResetEncoder(void);
	void		Encode(void);

	int		usedBits;					/* # bits filled in current byte */
	ULong		currentByte;				/* current byte */

	size_t	initialCodeSize;
	size_t	currentCodeSize;
	ULong		clearCode;
	ULong		endCode;
	ULong		newCodes;
	ULong		topSlot;
	ULong		slot;
	long		hashTable[HASH_SIZE];
	UShort	codeTable[HASH_SIZE];

// GIF output stream
	void		WriteHeader(void);
	void		WriteLogicalScreenDescriptor(void);
	void		WriteGlobalColorTable(void);
	void		WriteGraphicControlExtension(void);
	void		WriteImageDescriptor(void);
	void		WriteImage(void);
	void		WriteTrailer(void);

	void		PutByte(UChar byte);
	void		PutWord(UShort word);
	void		PutCode(int code);
	void		PutBlockByte(UChar byte);
	void		FlushBlock(void);

	UChar *	dataPtr;
	UChar *	dataStart;
	UChar *	dataStop;

	int		blockSize;
	UChar		block[256];
};

	
#endif	/* __GIF_H */
